/*-
 * Copyright (c) 2002 Doug White
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * 
 */

/*
 * kcs-common.c
 * Common routines for IPMI KCS interface
 */
#include <sys/types.h>
#include <machine/cpufunc.h>
#include <machine/sysarch.h>
#include <stdio.h>
#include "kcs.h"

u_int8_t
kcs_get_state()
{
    return inb(KCS_STATUS_IOPORT);
}

u_int8_t
kcs_print_state(u_int8_t status)
{
    /* we assume we have already ioperm'd the space */
    printf("Current KCS state: 0x%x : ", status);
    if((status & KCS_STATBITS) == KCS_IDLE) {
            printf("IDLE_STATE ");
    } else if((status & KCS_STATBITS) == KCS_READ) {
            printf("READ_STATE ");
    } else if((status & KCS_STATBITS) == KCS_WRITE) {
            printf("WRITE_STATE ");
    } else if((status & KCS_STATBITS) == KCS_ERROR) {
            printf("ERROR_STATE ");
    } else {
            printf("Huh? ");
    }
    if(status & KCS_IBF) {
            printf("IBF ");
    }
    if(status & KCS_OBF) {
            printf("OBF ");
    }
    if(status & KCS_OEM1) {
            printf("OEM1 ");
    }
    if(status & KCS_OEM2) {
            printf("OEM2 ");
    }
    printf("\n");
    return status;
}

/*
 * Wait for IBF (In-Bound Flag) to clear, signalling BMC has
 * read the command. 
 */
void
kcs_wait_for_ibf_clear()
{
    while(kcs_get_state() & KCS_IBF)
    {
#ifdef DEBUG
        printf("SLEEPING ");
#endif
        usleep(KCS_SLEEP_USECS);
    }
}

/* 
 * Wait for OBF to raise, signalling data is pending to read
 * or no command is pending.
 */

void
kcs_wait_for_obf_set()
{
    while(!(kcs_get_state() & KCS_OBF))
    {
#ifdef DEBUG
        printf("SLEEPING ");
#endif
        usleep(KCS_SLEEP_USECS);
    }
}

/*
 * Read byte from outbound data port. 
 */
u_int8_t
kcs_read_byte()
{
   return inb(KCS_DATAOUT_IOPORT);
}

/*
 * Bump channel into sending next byte.
 */
void
kcs_read_next() 
{
    outb(KCS_DATAIN_IOPORT, KCS_CMD_READ);
}
/*
 * Set up channel for writing.
 */
void
kcs_start_write()
{
    outb(KCS_CMD_IOPORT, KCS_CMD_WRITE_START);
}

/*
 * Write byte to inbound data port.
 */
void
kcs_write_byte(u_int8_t byte)
{
    outb(KCS_DATAIN_IOPORT, byte);
}

/* 
 * Set up channel to end write.
 */
void
kcs_end_write()
{
    outb(KCS_CMD_IOPORT, KCS_CMD_WRITE_END);
}

/* 
 * Send Abort current processing command.
 */
void
kcs_abort()
{
    outb(KCS_CMD_IOPORT, KCS_CMD_ABORT);
}

/*
 * XXX Could shorten this to 
 * return ((kcs_get_state() & KCS_STATBITS) == state)
 */
int
kcs_test_if_state(u_int8_t state)
{
#ifdef DEBUG
    kcs_print_state(kcs_get_state());
    printf("Checking STATBITS 0x%x against requested 0x%x\n", (kcs_get_state() & KCS_STATBITS), state & KCS_STATBITS);
#endif
    if((kcs_get_state() & KCS_STATBITS) == (state & KCS_STATBITS))
    {
#ifdef DEBUG
	printf("kcs_get_state() matched\n");
#endif
        return 1;
    }
    else
    {
#ifdef DEBUG
	printf("kcs_get_state() did not match\n");
#endif
        return 0;
    }
}

/*
 * Read dummy byte to clear OBF if set.
 */
void
kcs_clear_obf()
{
    if(kcs_get_state() & KCS_OBF) 
    {
        kcs_read_byte();
    }
}

/* 
 * Main read loop.
 */
 
int
kcs_read(u_int8_t* data, int len)
{
    u_int8_t *p = data;
#ifdef DEBUG
    u_int8_t x;
#endif
    int l=1; /* be nice about final position */
    if(data == NULL) 
    {
        return -1;
    }
    kcs_wait_for_ibf_clear();
    if(!kcs_test_if_state(KCS_READ)) 
    {
        printf("Not in READ state?\n");
        exit(2);
    }
    while(kcs_test_if_state(KCS_READ) && l <= len)
    {
        kcs_wait_for_obf_set();
#ifdef DEBUG
        x = kcs_read_byte();
        printf("Read byte 0x%x\n", x);
        *(p++) = x;
#else 
        *(p++) = kcs_read_byte();
#endif
        l++;
        kcs_read_next();
        kcs_wait_for_ibf_clear();
    }
    if(kcs_test_if_state(KCS_IDLE))
    {
        /* Clean up */
        kcs_wait_for_obf_set();
        kcs_read_byte(); /* toss it, ACK */
        return l;
    }
    else
    {
        /* error! */
        return -1;
    }
}

/*
 * Standard write loop. 
 */
int
kcs_write(u_int8_t *data, int len)
{
    u_int8_t *p=data;
    int l=len;
    
    kcs_wait_for_ibf_clear();
    kcs_clear_obf();
    kcs_start_write();
    kcs_wait_for_ibf_clear();
    if(!kcs_test_if_state(KCS_WRITE))
    {
        printf("Couldn't get into WRITE state\n"); 
        return -1;
    }
    kcs_clear_obf();

    /* 
     * Main data shlep loop.
     * pointer arith is fun. 
     * note we have to save last byte.
     */
#ifdef DEBUG
    printf("Writing %d bytes - 0x%x , 0x%x\n", len, p[0], p[1]);
#endif
    /* for(p=data; data+len-1 < p; p++) */
    for(; len > 1; len--)
    {
#ifdef DEBUG
        printf("Writing byte 0x%x\n", *p);
#endif
        kcs_write_byte(*p);
        kcs_wait_for_ibf_clear();
        if(!kcs_test_if_state(KCS_WRITE))
        {
            printf("Got out of WRITE state in write loop\n");
            return -1;
        }
        kcs_clear_obf();
        p++;
    }
    kcs_end_write();
    kcs_wait_for_ibf_clear();
    if(!kcs_test_if_state(KCS_WRITE))
    {
        printf("Got out of write state when done\n");
        return -1;
    }
#ifdef DEBUG
    printf("After final WRITE_STATE test:\n");
    kcs_print_state(kcs_get_state());
#endif
    kcs_clear_obf();
    kcs_write_byte(*p);
/*    if(!kcs_test_if_state(KCS_READ)) {
       printf("Not in READ state after writing last byte?\n");
       kcs_print_state(kcs_get_state());
       exit(1);
    }
*/
    return len;
}
    


