/*-
 * Copyright (c) 2002 Doug White
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * 
 */

/*
 * ioportmodule.c
 * Python/C module to access inb()/outb() and system I/O ports
 *
 * Doug White
 */

#include "Python.h"
#include <sys/types.h>
#include <machine/cpufunc.h>
#include <machine/sysarch.h>


/* 
 * ioport_ioperm: Get permission from the OS to read & write from
 * the I/O ports.
 *
 * Args:
 * . starting address
 * . length
 * . boolean activate/deactivate range
 */

PyObject *
ioport_ioperm(PyObject* self, PyObject* args)
{
        unsigned int start;
        unsigned int length;
        int enable;

        int t;
                                
        if (!PyArg_ParseTuple(args, "iii:ioport_ioperm", &start, &length, &enable))
        {
                return NULL;
        }

        /* Make the call */

        if(i386_set_ioperm(start, length, enable) == -1)
        {
                /* boo boo */
                PyErr_SetFromErrno(PyExc_OSError);
                return NULL;
        }

        Py_INCREF(Py_None);
        return Py_None;
}

/* inb()
 * Read byte from I/O port
 */

PyObject*
ioport_inb(PyObject* self, PyObject* args)
{
        int addr;
        int value;

        if(!PyArg_ParseTuple(args, "i:ioport_inb", &addr)) 
        {
                return NULL;
        }
                        
        /* Make the call */

        value = inb(addr);

        return Py_BuildValue("i", value);
}

/* 
 * outb()
 * Write byte to I/O port
 */
          
PyObject *
ioport_outb(PyObject* self, PyObject* args)
{
        int addr;
        int value;

        if(!PyArg_ParseTuple(args, "ii:ioport_outb", &addr, &value))
        {
                return NULL;
        }

        outb(addr, value);

        Py_INCREF(Py_None);
        return Py_None;
}

/* Method Table */

static PyMethodDef IoportMethods[] = {
        {"ioperm", ioport_ioperm, METH_VARARGS, "Set I/O permissions"},
        {"inb", ioport_inb, METH_VARARGS, "Read byte from I/O space"},
        {"outb", ioport_outb, METH_VARARGS, "Write byte to I/O space"},
        {NULL, NULL, 0, NULL}
};

void 
initioport(void)
{
        (void) Py_InitModule("ioport", IoportMethods);
}



