#
# Copyright (c) 2002 Doug White
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# $Id: ipmicommon.py,v 1.13 2002/07/20 02:31:01 dwhite Exp $
#
# ipmicommon -- Common functions for IPMI features
#
# Doug White

import exceptions
import md5
import socket
import struct
import time
import ipmiexception
import ipmimessage

# Some constants.
IPMI_AUTH_CALLBACK = 1
IPMI_AUTH_USER = 2
IPMI_AUTH_OPERATOR = 3
IPMI_AUTH_ADMIN = 4
IPMI_AUTH_OEM = 5

IPMI_AUTHT_NONE = 0
IPMI_AUTHT_MD2 = 1
IPMI_AUTHT_MD5 = 2
IPMI_AUTHT_PASSWD = 4
IPMI_AUTHT_OEM = 5

# IPMI Session: 
# Abstraction for handling IPMI Sessions, which maintain
# connection-oriented data.
class IPMISession:
    def __init__(self, peer, username='', password='', authlevel=IPMI_AUTH_ADMIN):
        # Create object with target pointed at peer
        self.peer = peer
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.bind(self.sock.getsockname())
        self.sessid = 0
        self.seq = 0
        self.iseq = 1
        self.username = username
        self.password = password
        self.authtypes = []
        self.authlevel = authlevel
        self.authtype = None
        self.authcode = None
        self.connected = 0

    # Calculate IPMI checksum.
    def calc_ipmi_cksum(self, body):
        # This is calculated by:
        # 2's complement cksum of preceeding bytes
        # 8 bit cksum
        # for byte in bytes:
        #  cksum = (cksum+byte) % 256
        # cksum = -cksum
        # then cksum + newcksum == 0
        cksum = 0
        for x in body:
            cksum = (cksum+ord(x)) % 256
        return -cksum
    
    def build_ipmi_channel_header(self, authtype, seq, sid, mac, leng):
        # All multibyte IPMI data is sent in Intel byte order, e.g. little endian.
        # No need to use hton* here. (bad Intel, no cookie)
        # byte 1: auth type
        # byte 2: sequence # (4 bytes)
        # byte 6: session ID (4 bytes)
        # byte 10: MAC (16 bytes)
        # there's a 1 byte Length after all this but we don't need that
        # MAC codes defined as:
        # 0 = none (when None is set AUTHCODE FIELD IS DROPPED)
        # 1 = MD2
        # 2 = MD5
        # 3 = reserved
        # 4 = straight password/key
        # 5 = Proprietary
                 
        # sequence number
        if seq == None:
            seq = 0

        # session ID
        if sid == None:
            sid = 0
                
        # Auth type
        if authtype == None:
            authtype = 0
            return struct.pack("<BLLB", authtype, seq, sid, leng)
        else:
            return struct.pack("<BLL16sB", authtype, seq, sid, mac, leng)

    # Construct IPMI source header.
    def build_ipmi_src_header(self, rqaddr, rqlun, rqseq, cmd):
        return struct.pack("<3b", rqaddr, ((rqseq << 2) + rqlun), cmd)

    # Construct IPMI destination header.
    def build_ipmi_dest_header(self, rsaddr, netfn, lun):
        # we have to combine these with their own checksum.
        nl = netfn << 2
        nl = nl + lun
        cksum = self.calc_ipmi_cksum("%c%c" % (rsaddr, nl))
        return struct.pack("<3b", rsaddr, nl, cksum)

    # calc_md5_authcode: Calculate MD5 auth code for insertion or comparison.
    def calc_md5_authcode(self, password, sessid, message, seq):
        # for MD* auth, we composite
        # password, session ID, ipmimessage, seq#, password
        # password must be 0-padded to 16 bytes
        m = md5.new()
        m.update(struct.pack("<16s", password))
        m.update(struct.pack("<L", sessid))
        m.update(message)
        m.update(struct.pack("<L", seq))
        m.update(struct.pack("<16s", password))
        return m.digest()

    # send_ipmi_command: Send IPMI command (including all headers and body)
    # to the configured peer. Slap in auth info if available.
    def send_ipmi_message(self, ipmimessage):
        # Calculate the AuthCode
        if self.authtype == IPMI_AUTHT_MD5:
            self.authcode = self.calc_md5_authcode(self.password, self.sessid, ipmimessage, self.seq)
        elif self.authtype == IPMI_AUTHT_MD2:
            raise ValueError, "MD2 Authentication is not supported"
        elif self.authtype == IPMI_AUTHT_PASSWD:
            # for password auth, just send the password (NO NO NO!!)
            self.authcode = self.password

        # Get the IPMI channel header, with message length.
        ipmisess = self.build_ipmi_channel_header(self.authtype, self.seq, self.sessid, self.authcode, len(ipmimessage))

        # Glue in RMCP header.
        # XXX this should use hton* and be put in __init__
        # since it's static
        rmcphead = struct.pack('<4b', 0x06, 0x00, 0xff, 0x07)

        # send!
        self.sock.sendto(rmcphead + ipmisess + ipmimessage, (self.peer, 0x26f))

        # Increment seq#
        if self.seq != 0 and self.seq != None:
            self.seq = self.seq + 1

        # Grab the response and return.
        return self.get_ipmi_response()

    # Extract the IPMI response information.
    def get_ipmi_body(self, resp):
        # first four bytes are the RMCP header, we can chuck 'em.
        # Next 8 are the session header
        (authtype, seq, sid) = struct.unpack("<BLL", resp[4:13])

        #if authtype != IPMI_AUTHT_NONE and authtype != IPMI_AUTHT_MD2:
        if authtype == IPMI_AUTHT_MD5:
            # pull off the MAC (16 bytes)
            mac = resp[13:29]
            # XXX must validate packet is authentic
            rmac = self.calc_md5_authcode(self.password, sid, resp[30:], seq)
            if rmac != mac:
                #print "MAC mismatch? %s VS %s" % (`mac`, `rmac`)
                raise ipmiexception.IPMIAuthenticationError, "Invalid AuthCode in packet"
            # next is length (what nut put this here? I smell overflow!)
            # leng = ord(resp[29])
            # rest should be the message body, unless that stupid pad byte
            # is hiding in there.
            return resp[30:]
        else:
            # leng = ord(resp[13])
            return resp[14:]

    # get_ipmi_response: Read packet from network, extract IPMI message,
    # validate authcode and return IPMIMessage.
    def get_ipmi_response(self):
        # Pull packet off wire
        resp = self.sock.recvfrom(100)[0]

        # Extract body and verify authcode
        resp = self.get_ipmi_body(resp)

        m = ipmimessage.IPMIMessage()
        m.unwrap(resp, 'LAN')
        return m
        

    # Extract the actual response data from the IPMI/IPMB packet.
    def get_response_data(self, body):
        (rqaddr, nl, cksum, rsaddr, rqsl) = struct.unpack("<5B", body[:5])
        return body[6:]

    # big-ass macro for building & sending ipmi data
    def send(self, daddr, dlun, saddr, slun, sseq, netfn, cmd, data):
        print "*** Something is using send()!  send() is deprecated! ***"
        header = self.build_ipmi_dest_header(daddr, netfn, dlun)
        body = self.build_ipmi_src_header(saddr, slun, sseq, cmd)
        body = body + data
        bck = self.calc_ipmi_cksum(body)
        ipmimessage = header + body + struct.pack("<b", bck)
        ret = self.send_ipmi_message(ipmimessage)
        nret = ret.data
        if ord(nret[0]) != 0:
            raise ipmiexception.IPMICommandError, ord(nret[0])
        else:
            return nret

    # IPMIMessage conformant version
    # If this message is going off-BMC, encapsulate.
    def sendmessage(self, message, chan=0):
        if message.rssa != 0x20:
            outer = ipmimessage.IPMIMessage()
            # XXX what should we use as a src address?!?
            # chan plus the 'bridging' bit since we're on LAN
            chan = chan + (1<<6)
            outer.quickload(0x20, 0x0, 0x20, 0x0, self.next_seq(), 0x6, 0x34, struct.pack('<B', chan) + str(message))
            ret = self.send_ipmi_message(str(outer))
            if ord(ret.data[0]) != 0:
                raise ipmiexception.IPMICommandError, ord(ret.data[0])
            # ret is the return from the Send Message command -- toss it
            # we need to grab the *next* message
            ret = self.get_ipmi_response()
        else:
            ret = self.send_ipmi_message(str(message))
            
        if ord(ret.data[0]) != 0:
            raise ipmiexception.IPMICommandError, ord(ret.data[0])
        
        return ret
        

    #####################################################
    #  COMMAND IMPLEMENTATIONS
    #####################################################
        
    # Determine auth capabilities of the peer.
    def get_auth_cap(self):
        # request data is two bytes, first is the channel num (0xe == this),
        # second is the max priv level (0x4=admin)
        body = struct.pack("<2b", 0x0e, self.authlevel)
        m = ipmimessage.IPMIMessage()
        m.quickload(0x20, 0x0, 0x81, 0x0, 0x0, 0x6, 0x38, body)
        rm = self.sendmessage(m)
        rbody = rm.data

        # now we have our response, parse the puppy out.
        # This is in the Section 12.4 format, with the IPMB header
        # still stuffed on it.
        # oemiddata is the 3 byte OEM ID plus 1 byte of data
        (ccode, chan, auths, aflags, reserved, oemiddata) = struct.unpack("<5bl", rbody)
        
        if auths & 0x1 != 0:
            self.authtypes.append('None')
        if auths & 0x2 != 0:
            self.authtypes.append('MD2')
        if auths & 0x4 != 0:
            self.authtypes.append('MD5')
        # 0x8 is reserved
        if auths & 0x10 != 0:
            self.authtypes.append('Password')
        if auths & 0x20 != 0:
            self.authtypes.append('OEM')

    # Get initial session challenge with requested authtype
    def get_session_challenge(self, atype):
        body = struct.pack("1B16s", atype, self.username)
        m = ipmimessage.IPMIMessage()
        m.quickload(0x20, 0x0, 0x81, 0x0, 0x0, 0x6, 0x39, body)
        rm = self.sendmessage(m)
        rbody = rm.data
        (ccode, self.sessid, self.authcode) = struct.unpack("<BL16s", rbody[:21])
        # future packets must be marked with this
        self.authtype = atype

    # Activate!
    def activate_session(self):
        # The fields for this get kinda tricky.  We start using the session
        # fields.  authtype is what we've been using; seq# is still 0;
        # sessid is the temporary one from GSC; we calcluate the authcode
        # for this message.
        # In the body, we re-pass the authtype, the max priv level,
        # the original challenge, and a remote ISN.
        # Activate Session is App/0x3a
        # self.authcode is still the challenge at this point
        # final byte is our ISN, probably should be random
        body = struct.pack("<2B16sL", self.authtype, self.authlevel, self.authcode, 0x1)
        m = ipmimessage.IPMIMessage()
        m.quickload(0x20, 0x0, 0x81, 0x0, 0x0, 0x6, 0x3a, body)
        rm = self.sendmessage(m)
        rbody = rm.data

        # well, if this worked, we should have something like this:
        try:
            (ccode, atype, self.sessid, self.seq, maxpriv) = struct.unpack("<2B2LB", rbody[:21])
        except:
            print "Error activating session (0x%x)" % ord(rbody[0])
            raise

    # Set authorization level. Since we start up in User, we need
    # to bump it up before doing anything else.
    def set_auth_level(self):
        m = ipmimessage.IPMIMessage()
        m.quickload(0x20, 0x0, 0x81, 0x0, 41, 0x6, 0x3b, struct.pack('<B', 4))
        rm = self.sendmessage(m)
        ret = rm.data
        if ord(ret[0]) != 0:
            raise ValueError, "Couldn't raise auth level"

    # Establish session with peer.
    def establish(self):
        # Bind our port.
        # Before we can bring the session up, we need to ask the host
        # what authentication is required/available.
        # get_auth_cap will load self.authtypes with the available
        # authentication methods.
        self.get_auth_cap()

        # Pick the best one.
        if 'MD5' in self.authtypes:
            atype = IPMI_AUTHT_MD5
        elif 'MD2' in self.authtypes:
            atype = IPMI_AUTHT_MD2
            print "*** MD2 Authentication is not supported! ***"
        elif 'Password' in self.authtypes:
            atype = IPMI_AUTHT_PASSWD
            print "*** DANGER: Using password auth ***"
            print "*** THE PASSWORD WILL BE SENT IN THE CLEAR! ***"
        else:
            raise exceptions.ValueError, "No supported auth types returned"

        # Get the initial session challenge.
        self.get_session_challenge(atype)

        # Activate session.
        self.activate_session()

        # Reset auth level.
        self.set_auth_level()

        # We are on-line!
        self.connected = 1

    def next_seq(self):
        self.iseq = self.iseq + 1
        return self.iseq
        
class IPMISystemInterface:
    def __init__(self, interface="SMIC"):
        # XXX
        import sys
        sys.path.append('ioport')
        import string
        import ioport

        self.interface = string.upper(interface)
        self.sseq = 1
        self.iseq = 1

        if self.interface not in ("KCS", "SMIC"):
            raise ValueError, "Not a valid interface"

        if self.interface == "KCS":
            import ipmikcs
            ioport.ioperm(ipmikcs.KCS_BASE_IOPORT, 2, 1)

        elif self.interface == "SMIC":
            import ipmismic
            ioport.ioperm(ipmismic.SMIC_BASE_PORT, 3, 1)

        else:
            print "huh?"

    def send_ipmi_kcs(self, tx):
        import ipmikcs
        ipmikcs.kcs_write(tx)
        ret = ipmikcs.kcs_read()[2:]
        if ord(ret[0]) != 0:
        #ret = ipmikcs.kcs_read()
        #if ord(ret[2]) != 0:
            raise ipmiexception.IPMICommandError, ord(ret[0])
        else:
            return ret

    def send_ipmi_smic(self, tx):
        import ipmismic
        ipmismic.smic_write(tx)
        ret = ipmismic.smic_read()[2:]
        if ord(ret[0]) != 0:
        #ret = ipmismic.smic_read()
        #if ord(ret[2]) != 0:
            raise ipmiexception.IPMICommandError, ord(ret[0])
        else:
            return ret


    # Get message off of Event Queue
    def getmessage(self):
        m = ipmimessage.IPMIMessage()
        m.quickload(0x20, 0, 0x20, 0, self.iseq, 0x6, 0x33, '')
        self.iseq = self.iseq + 1
        ret = self.sendmessage(m)
        if ret and ord(ret.data[0]) == 0x80:
            return None
        else:
            return ret

    # IPMISession-compatible send
    def send(self, daddr, dlun, saddr, slun, sseq, netfn, cmd, data):
        print "*** Something is using send()!  send() is deprecated! ***"
        # in SI land, we only need the netfn, command, and data
        tx = ""
        tx = struct.pack('<BB', ((netfn << 2) + slun), cmd)
        tx = tx + data
        # Increment stored sequence number for use externally
        self.sseq = self.sseq + 1

        if self.interface == "KCS":
            return self.send_ipmi_kcs(tx)

        elif self.interface == "SMIC":
            return self.send_ipmi_smic(tx)

        else:
            raise ipmiexception.IPMIError, "No such interface type"

    def sendmessage(self, message, chan=0):
        offbmc = 0
        ret = ''
        if message.rssa != 0x20:
            offbmc = 1
            # flag response messages to come to the EMQ
            message.rqsa = 0x20
            message.rqlun = 0x2
            outer = ipmimessage.IPMIMessage()
            # XXX what should we use as a src address?!?
            outer.quickload(0x20, 0x0, 0x20, 0x0, self.next_seq(), 0x6, 0x34, struct.pack('<B', chan) + str(message))
            message = outer
            
        tx = ""
        tx = struct.pack('<BB', ((message.netfn << 2) + message.rslun), message.cmd)
        tx = tx + message.data

        if self.interface == "KCS":
            ret = self.send_ipmi_kcs(tx)

        elif self.interface == "SMIC":
            ret = self.send_ipmi_smic(tx)

        else:
            raise ipmiexception.IPMIError, "No such interface type"

        m = ipmimessage.IPMIMessage()
        if offbmc:
            # Busy wait on the event message queue to get the
            # response from the bridged request.
            ret = None
            while not ret:
                try:
                    time.sleep(0)
                    ret = self.getmessage()
                except ipmiexception.IPMICommandError, e:
                    if e.value == 0x80:
                        pass

            m.unwrap(ret.data[2:], 'IPMB')
            
        else:
            m.data = ret
        return m

    def establish(self):
        pass

    def next_seq(self):
        self.iseq = self.iseq + 1
        return self.iseq

