#
# Copyright (c) 2002 Doug White
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# $Id: ipmifru.py,v 1.3 2002/07/14 21:50:00 dwhite Exp $
#
# ipmifru.py: Methods for manipulating the FRU repository.
#
# Doug White

import struct
import time
import ipmitransport

# pbytes: debugging print
def pbytes(ret):
    for y in ret:
        print "0x%x ," % ord(y)

# FRU access notes:
# . Every EEPROM has an 8 byte Common Header which gives offset pointers
#   to other regions in the EEPROM, in 8-byte multiples(!).
# . Each area has a version tag.
# . Each field has an encoding & length tag prefix
# . Encodings are BCD, 6-bit ascii, or 8-bit ASCII.
# . Each EERPOM has up to 5 regions:
#   . Internal Use Area: private for BMC
#   . Chassis Info Area: type, p/n and s/n
#   . Board Info Area: mfr stamp, board mfr p/n s/n, custom mfr fields
#   . Product Info Area: mfr name, part/serial/product name, asset tag
#   . MultiRecord Area: Various types (power, mgmt fields, some from DMTF)

# decode_chassis_type: Convert numeric chassis type into text.

chassistypes = ['Reserved', 'Other', 'Unknown', 'Desktop',
                'Low Profile Desktop', 'Pizza Box', 'Mini Tower', 'Tower',
                'Portable', 'LapTop', 'Notebook', 'Hand Held',
                'Docking Station', 'All in One', 'Sub Notebook',
                'Space-saving', 'Lunch Box', 'Main Server Chassis' ,
                'Expansion Chassis', 'SubChassis', 'Bus Expansion Chassis',
                'Peripheral Chassis', 'RAID Chassis', 'Rack Mount Chassis']

def decode_chassis_type(ct):
    if ct < len(chassistypes):
        return chassistypes[ct]
    else:
        return "Undefined"

# parse_fru_fields: split up TLV-encoded values in FRU tables

def parse_fru_fields(data):
    strs = []
    while len(data) > 1:
        # (str, dat) = parse_fru_field(dat)
        if ord(data[0]) == 0xc1:
            break
        ftype = (ord(data[0]) & 0xc0) >> 6
        flen = ord(data[0]) & 0x3f
        if ftype != 0x2 and ftype != 0x3:
            # print "Unknown field type 0x%x" % ftype
            continue
        field = data[1:flen+1]
        strs.append(field)
        data = data[flen+1:]
    return strs

# get_fru_data: Retrieve the data from a given FRU SDR. Must give the SDR
# record of the FRU to retrieve data from.

def get_fru_data(sess, fsdr):

    # the Read FRU Data command needs the device ID, two-byte offset, and
    # the number of bytes to read, kinda like chunked records.

    rec = {}

    if fsdr['device-access'] == 0x0:
        # Directly on IPMB; device-address >> 1 is actual slave location
        # but we don't implement Master Read/Write yet.
        raise ValueError, "IPMB-Connected devices not supported yet"

    # If this is a DIMM device, decode it specially.
    if fsdr['device-type-code'] == 0x10 and fsdr['device-type-modifier'] == 0x1:
        ret = ipmitransport.load_fru_chunked_record(sess, 0x81, 0x0, 1, 0xa, 0x11, fsdr, 0, 1)
        spdlen = ord(ret[0])
        print "SPD Length = %d" % spdlen
        # now get the rest
        ret = ipmitransport.load_fru_chunked_record(sess, 0x81, 0x0, 1, 0xa, 0x11, fsdr, 0, spdlen)
        # this assumes native endianness
        rec['spd-size'] = 2**ord(ret[1])
        rec['fundamental-type'] = ord(ret[2])
        if ord(ret[2]) == 0x2:
            rec['fundamental-type-name'] = "EDO"
        elif ord(ret[2]) == 0x4:
            rec['fundamental-type-name'] = "SDRAM"
        rec['number-of-rows'] = ord(ret[3])
        rec['number-of-columns'] = ord(ret[4])
        rec['number-module-rows'] = ord(ret[5])
        #rec['data-width'] = ret[6:8]
        rec['data-width'] = (ord(ret[7]) << 8) + ord(ret[6])
        rec['voltage-interface'] = ord(ret[8])
        if ord(ret[8]) == 0x0:
            rec['voltage-interface-name'] = "5.0 Volt/TTL"
        elif ord(ret[8]) == 0x1:
            rec['voltage-interface-name'] = "LVTTL"
        elif ord(ret[8]) == 0x2:
            rec['voltage-interface-name'] = "HSTL 1.5"
        elif ord(ret[8]) == 0x3:
            rec['voltage-interface-name'] = "SSTL 3.3"
        elif ord(ret[8]) == 0x4:
            rec['voltage-interface-name'] = "SSTL 2.5"
        rec['cycle-time-highest-cas'] = ord(ret[9])
        rec['access-from-clock-highest-cas'] = ord(ret[10])
        rec['dimm-config-type'] = ord(ret[11])
        if ord(ret[11]) == 0x0:
            rec['dimm-config-type-name'] = "None"
        elif ord(ret[11]) == 0x1:
            rec['dimm-config-type-name'] = "Parity"
        elif ord(ret[11]) == 0x2:
            rec['dimm-config-type-name'] = "ECC"
        rec['refresh-rate'] = ord(ret[12])
        rec['primary-sdram-width'] = ord(ret[13])
        rec['error-checking-sdram-width'] = ord(ret[14])
        rec['min-clock-delay'] = ord(ret[15])
        rec['burst-lengths-supported'] = ord(ret[16])
        rec['number-of-banks'] = ord(ret[17])
        rec['cas-latencies-supported'] = ord(ret[18])
        rec['cs-latency'] = ord(ret[19])
        rec['write-latency'] = ord(ret[20])
        rec['sdram-module-attributes'] = ord(ret[21])
        rec['sdram-device-attributes'] = ord(ret[22])
        rec['cycle-time-second-cas'] = ord(ret[23])
        rec['access-from-clock-second-cas'] = ord(ret[24])
        rec['cycle-time-third-cas'] = ord(ret[25])
        rec['access-from-clock-third-cas'] = ord(ret[26])
        rec['min-row-precharge'] = ord(ret[27])
        rec['min-active-to-active'] = ord(ret[28])
        rec['min-ras-to-cas'] = ord(ret[29])
        rec['min-ras-pulse-width'] = ord(ret[30])
        rec['row-density'] = ord(ret[31])
        rec['command-signal-setup-time'] = ord(ret[32])
        rec['command-signal-hold-time'] = ord(ret[33])
        rec['data-signal-setup-time'] = ord(ret[34])
        rec['data-signal-hold-time'] = ord(ret[35])
        rec['superset'] = ret[36:62]
        rec['spd-data-revision'] = ord(ret[62])
        rec['checksum'] = ord(ret[63])
        rec['mfr-code'] = ret[64:72]
        rec['mfr-location'] = ord(ret[72])
        rec['mfr-part-number'] = ret[73-91]
        rec['revision-code'] = ret[91:93]
        rec['mfr-date'] = ret[93:95]
        rec['serial-number'] = ret[95:99]
        rec['mfr-data'] = ret[99:126]
        rec['intel-spec-frequency'] = ord(ret[126])
        rec['intel-spec-cas-latency'] = ord(ret[127])
        k = rec.keys()
        k.sort()
        for l in k:
            if l == "mfr-code":
                print "%s: %s" % (l, `rec[l]`)
            else:
                print "%s: %s" % (l, rec[l])
        return rec

    # else...

    # Grab the Common Header (8 bytes)
    ret = ipmitransport.load_fru_chunked_record(sess, 0x81, 0x0, 1, 0xa, 0x11, fsdr, 0, 8)
    chfver, iuaoff, ciaoff, baoff, piaoff, mraoff, pad, chksum = struct.unpack('<8B', ret)
##    print "Common Header Version: 0x%x" % chfver
##    if iuaoff:
##        print "Internal Use Area offset: 0x%x" % iuaoff
##    if ciaoff:
##        print "Chassis Info Area offset: 0x%x" % ciaoff
##    if baoff:
##        print "Board Info Area offset: 0x%x" % baoff
##    if piaoff:
##        print "Product Info Area offset: 0x%x" % piaoff
##    if mraoff:
##        print "MultiRecord Area offset: 0x%x" % mraoff

    # check out chassis area first
    # note that offset needs to be requested as byte offset but is *given*
    # as n*8 byte offset (i.e., << 3)
    if ciaoff:
        off = ciaoff * 8

        # Pull down Chassis Info Area header
        ret = ipmitransport.load_fru_chunked_record(sess, 0x81, 0x0, 1, 0xa, 0x11, fsdr, off, 3)
        ciaver, cialen, chastype = struct.unpack('<3B', ret)
        cialen = cialen * 8

##        print "Chassis Area version: 0x%x" % ciaver
##        print "Chassis Area length: %d bytes" % cialen
##        print "Chassis Type: 0x%x" % chastype

        ret = ipmitransport.load_fru_chunked_record(sess, 0x81, 0x0, 1, 0xa, 0x11, fsdr, off+3, cialen-3)
        fields = parse_fru_fields(ret)
        #print fields
        rec['chassis-type'] = chastype
        rec['chassis-type-name'] = decode_chassis_type(chastype)
        rec['chassis-part-number'] = fields[0]
        rec['chassis-serial-number'] = fields[1]

    if baoff:
        off = baoff * 8

        # Pull down Board Info Area header
        ret = ipmitransport.load_fru_chunked_record(sess, 0x81, 0x0, 1, 0xa, 0x11, fsdr, off, 6)
        baver, balen, balang, mfgstamp = struct.unpack('<3B3s', ret)
        balen = balen * 8

        # MFG time is minutes since 00:00 Jan 1, 1996
        # == 820454400 / 60 = 13674240
        mfgmin = 13674240+ ord(mfgstamp[0]) + (ord(mfgstamp[1]) << 8) + (ord(mfgstamp[2]) << 16)

##        print "Board Area version: 0x%x" % baver
##        print "Board Area length: %d bytes" % balen
##        print "Board Area language: 0x%x" % balang
##        print "Board Mfg time: %s" % time.ctime(mfgmin*60)

        ret = ipmitransport.load_fru_chunked_record(sess, 0x81, 0x0, 1, 0xa, 0x11, fsdr, off+6, balen-6)
        fields = parse_fru_fields(ret)
        # print fields
        rec['board-manufactured'] = mfgmin*60  # UNIX timestamp
        rec['board-manufacturer'] = fields[0]
        rec['board-product-name'] = fields[1]
        rec['board-serial-number'] = fields[2]
        rec['board-part-number'] = fields[3]

    if piaoff:
        off = piaoff * 8

        # Pull down Product Info Area header
        ret = ipmitransport.load_fru_chunked_record(sess, 0x81, 0x0, 1, 0xa, 0x11, fsdr, off, 3)
        piaver, pialen, pialang = struct.unpack('<3B', ret)
        pialen = pialen * 8

##        print "Product Area version: 0x%x" % piaver
##        print "Product Area length: %d bytes" % pialen
##        print "Product Area language: 0x%x" % pialang

        ret = ipmitransport.load_fru_chunked_record(sess, 0x81, 0x0, 1, 0xa, 0x11, fsdr, off+3, pialen-3)
        fields = parse_fru_fields(ret)
        # print fields
        rec['product-manufacturer'] = fields[0]
        rec['product-name'] = fields[1]
        rec['product-part-number'] = fields[2]
        rec['product-version'] = fields[3]
        rec['product-serial-number'] = fields[4]
        rec['product-asset-tag'] = fields[5]

    # MulitRecord table
    if mraoff:
##        print "MultiRecord record types:"
        # This is a bit different... it's a chain of records with header
        # separators that point you to the next record. Fun.
        
        off = mraoff * 8

        recver = 0
        while recver & 0x80 == 0:
            # The record headers are 5 bytes
            ret = ipmitransport.load_fru_chunked_record(sess, 0x81, 0x0, 1, 0xa, 0x11, fsdr, off, 5)
            rectype, recver, reclen, rck, hck = struct.unpack('<5B', ret)
            # grab body bytes
            off = off + 5
            ret = ipmitransport.load_fru_chunked_record(sess, 0x81, 0x0, 1, 0xa, 0x11, fsdr, off, reclen)
            off = off + reclen

            rtstr = "" # record type string
            rtspec = "" # record specific string
            
            if rectype == 0x0:
                rtstr = "power supply"
            elif rectype == 0x1:
                rtstr = "DC Output"
            elif rectype == 0x2:
                rtstr = "DC Load"
            elif rectype == 0x3:
                rtstr = "Management Access Record"
                # first byte of record has a subtype
                recsubtype = ord(ret[0])
                recdata = ret[3:]
                if recsubtype == 0x1:
                    rtspec = ("System Management URL", recdata)
                elif recsubtype == 0x2:
                    rtspec = ("System Name", recdata)
                elif recsubtype == 0x3:
                    rtspec = ("System Ping Address", recdata)
                elif recsubtype == 0x4:
                    rtspec = ("Component Management URL" , recdata)
                elif recsubtype == 0x5:
                    rtspec = ("Component Name", recdata)
                elif recsubtype == 0x6:
                    rtspec = ("Component Ping Address", recdata)
                elif recsubtype == 0x7:
                    rtspec = ("System Unique ID", recdata)
            elif rectype == 0x4:
                rtstr = "Base Compatibility Record"
            elif rectype == 0x5:
                rtstr = "Extended Compatibility Record"
            elif rectype >= 0xc0 and rectype < 0xff:
                rtstr = "OEM"
                pbytes(ret[2:])
            elif rectype == 0xff:
                rtstr = "Invalid data? 0xff means no such address, usually.."
            else:
                rtstr = "Unknown MultiRecord type 0x%x" % rectype

    return rec
