#
# Copyright (c) 2002 Doug White
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# $Id: ipmimessage.py,v 1.7 2002/07/20 02:32:25 dwhite Exp $
#
# IPMI message wrapping & unwrapping

import struct
import ipmiexception

class IPMIMessage:
    def __init__(self):
        self.netfn = 0
        self.cmd = 0
        self.rqsa = 0
        self.rqlun = 0
        self.rqseq = 0
        self.rssa = 0
        self.rslun = 0
        self.handle = 0
        self.data = ''
        self.hchk = ''
        self.bchk = ''

    def __str__(self):
        # build a full IPMB message
        header = self.build_ipmi_dest_header(self.rssa, self.netfn, self.rslun)
        body = self.build_ipmi_src_header(self.rqsa, self.rqlun, self.rqseq, self.cmd)
        body = body + self.data
        bck = self.calc_ipmi_cksum(body)
        ipmimessage = header + body + struct.pack("<b", bck)
        
        return ipmimessage

    def __repr__(self):
        return "<IPMIMessage object rssa=0x%x netfn=0x%x cmd=0x%x>" % (self.rssa, self.netfn, self.cmd)

    def dump(self):
        print repr(self)
        for v in vars(self).keys():
            if v == "self":
                continue
            elif v == "data":
                print "data: ", `self.data`
                continue
            print "%s: %s" % (v, getattr(self, v)) 

    # For whatever reason parsing bridged messages is a real bear since
    # the message format depends on where it came from.  The PCI people
    # got lucky, they had their extraneous fields zeroed.
    # The smallest common format is a message from IPMB received with the
    # Get Message command.  The longest is a message sent to a multichannel
    # type channel (like LAN); a session handle is prepended to the 6-byte
    # IPMB format.

    def unwrap(self, mess, source):
        if len(mess) < 6:
            raise ipmiexception.IPMIMessageError, "Not a valid IPMB message"

        # LAN with 1-byte session handle
        remain = ''
        try:
            if source == "S-LAN":
                self.handle, self.rssa, netfnlun, self.chk1, rqsa, rqseqlun, self.cmd = struct.unpack('<7B', mess[:7])
                remain = mess[7:]
            elif source == "LAN":
                # print "matched lan"
                self.rssa, netfnlun, self.chk1, rqsa, rqseqlun, self.cmd = struct.unpack('<6B', mess[:6])
                remain = mess[6:]
            elif source == "IPMB":
                # first 5 bytes
                netfnlun, self.chk1, rqsa, rqseqlun, self.cmd = struct.unpack('<5B', mess[:5])
                remain = mess[5:]
            else:  # assume IPMB
                 netfnlun, self.chk1, rqsa, rqseqlun, self.cmd = struct.unpack('<5B', mess[:5])
                 remain = mess[5:]
        except:
            raise ipmiexception.IPMIMessageError, "Problem unpacking"
        
        self.netfn = netfnlun >> 2
        self.rqseq = rqseqlun >> 2
        if self.netfn % 2:
            # response
            self.rqlun = netfnlun & 0x3
            self.rssa = rqsa
            self.rslun = rqseqlun & 0x3
        else:
            # request
            self.rslun = netfnlun & 0x3
            self.rqsa = rqsa
            self.rqlun = netfnlun & 0x3
            
        # check checksum?

        # XXX chk2 wil reappear later once get_ipmi_body() is fixed
        self.data = remain[:-1]
        self.chk2 = remain[-1]
        #self.data = remain

    # ipmicommon.send() compatible interface
    def quickload(self, daddr, dlun, saddr, slun, sseq, netfn, cmd, data):
        self.rqsa = saddr
        self.rqlun = slun
        self.rssa = daddr
        self.rslun = dlun
        self.rqseq = sseq
        self.netfn = netfn
        self.cmd = cmd
        self.data = data

    ##### the ugly part #####

    # Construct IPMI source header.
    def build_ipmi_src_header(self, rqaddr, rqlun, rqseq, cmd):
        return struct.pack("<3b", rqaddr, ((rqseq << 2) + rqlun), cmd)

    # Construct IPMI destination header.
    def build_ipmi_dest_header(self, rsaddr, netfn, lun):
        # we have to combine these with their own checksum.
        nl = netfn << 2
        nl = nl + lun
        cksum = self.calc_ipmi_cksum("%c%c" % (rsaddr, nl))
        return struct.pack("<3b", rsaddr, nl, cksum)

    # Calculate IPMI checksum.
    def calc_ipmi_cksum(self, body):
        # This is calculated by:
        # 2's complement cksum of preceeding bytes
        # 8 bit cksum
        # for byte in bytes:
        #  cksum = (cksum+byte) % 256
        # cksum = -cksum
        # then cksum + newcksum == 0
        cksum = 0
        for x in body:
            cksum = (cksum+ord(x)) % 256
        return -cksum
        

# Dequeue message from Request Message Queue (SI only)
def get_queued_message(sess):
    try:
        ret = sess.send(0x20, 0, 0x20, 0, 0, 0x6, 0x33, '')
    except ipmiexception.IPMICommandError, e:
        if e.value == 0x80:
            # no mail
            return None
        else:
            raise

    x = IPMIMessage()
    x.unwrap(ret)
    return x



