#
# Copyright (c) 2002 Doug White
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# $Id: ipmisdr.py,v 1.9 2002/07/19 03:30:40 dwhite Exp $
#
# ipmisdr.py: Methods for manipulating the SDR repository.
#
# Doug White

import struct
import ipmidevice
import ipmievent
import ipmimessage
import ipmitransport

# pbytes: debugging print
def pbytes(ret):
    for y in ret:
        print "0x%x ," % ord(y)

# utos - convert unsigned to singed integer with given width
def utos(val, bits):
    if val & (2**(bits-1)):
        return -((~val & 2**(bits-1)-1)+1)
    else:
        return val

# check-if-convert: Check if we need to do a conversion.
def check_if_format(format, val, bits):
    if format == 0 or format == 3:
        return val
    if format == 2:
        return utos(val, bits)
    elif format == 1:
        raise ValueError, "1's Complement is not supported yet"
    else:
        raise ValueError, "Odd value passed to check_if_format"

# generate-sdr-key: Generate a key for a dictionary of SDRs
# looks at the record type and generates an appropriate tuple.
def generate_sdr_key(dr):
    rt = dr['record-type']
    if rt == 0x1:
        return (dr['record-type'], dr['sensor-owner-id'], dr['sensor-owner-lun'], dr['sensor-number'])
    elif rt == 0x2:
        return (dr['record-type'], dr['sensor-owner-id'], dr['sensor-owner-lun'], dr['sensor-number'])
    elif rt == 0x10:
        return (dr['record-type'], dr['device-access'], dr['device-address'])
    elif rt == 0x11:
        # return (dr['record-type'], dr['device-access'], dr['device-address'], dr['access-info'], dr['channel-number'])
        return (dr['record-type'], dr['device-access'], dr['device-address'])
    else:
        # XXX bad!
        return "unsupported"

# decode_sdr_record: Take SDR entry (including the record header)
# and split into a dictionary of values from that record.

# The SDR repository defines a lot of fields :-)
# Use a dictionary to speed things along.

def decode_sdr_record(data):
    rec = {}

    # Extract the header first.
    ridlo, ridhi, sdrver, rtype, rlen = struct.unpack('<BBBBB', data[:5])

    # XXX
    #print "#%d: Type 0x%x, version 0x%x, len %d" % ((ridhi << 8)+ridlo, rtype, sdrver, rlen)

    rec['record-id'] = (ridhi<<8)+ridlo
    rec['record-type'] = rtype
    rec['sdr-version'] = sdrver
    rec['record-length'] = rlen

    # Now split off the rest of the parsing by type.

    if rtype == 0x1:
        try:
            sownid, solun, snum, entid, entinst, sinit, scap, stype, etype, amasklo, amaskhi, dmasklo, dmaskhi, dimasklo, dimaskhi, sunit1, sunit2, sunit3, lin, m, mtol, b, bacc, acc, rbexp, achar, nread, nmax, nmin, smax, smin, unonrecov, ucrit, unoncrit, lnonrecov, lcrit, lnoncrit, physt, nhyst, rev1, rev2, oem, idtype = struct.unpack('<BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB', data[5:48])
            idstr = data[48:]
        except:
            print "struct bombed, len = %d" % len(data[5:48])
            pbytes(data[5:48])
            raise

        rec['sensor-owner-id'] = sownid
        rec['sensor-owner-lun'] = solun & 0x3
        rec['sensor-owner-channel'] = (solun & 0xf0) >> 4
        rec['sensor-number'] = snum
        rec['entity-id'] = entid
        if entid < len(ipmievent.entitytypelist):
            rec['entity-id-name'] = ipmievent.entitytypelist[entid]
        rec['entity-is-logical'] = entinst & 0x80
        rec['entity-instance'] = entinst & 0x7f
        rec['sensor-initialization'] = sinit
        rec['sensor-capabilities'] = scap
        rec['sensor-capability-hysteresis'] = (scap & 0x30) >> 4
        rec['sensor-capability-thresh'] = (scap & 0xc) >> 2
        rec['sensor-capability-events'] = scap & 0x3              
        rec['sensor-type'] = stype
        if stype < len(ipmievent.sensortypelist):
            rec['sensor-type-name'] = ipmievent.sensortypelist[stype]
        rec['event-type'] = etype
        if etype < len(ipmievent.sensortypelist):
            rec['event-type-name'] = ipmievent.eventtypelist[etype]
        rec['assert-event-mask'] = (amaskhi << 8) + amasklo
        rec['lower-thresh-mask'] = (amaskhi << 8) + amasklo
        rec['deassert-event-mask'] = (dmaskhi << 8) + dmasklo
        rec['upper-thresh-mask'] = (dmaskhi << 8) + dmasklo
        rec['discrete-reading-mask'] = (dimaskhi << 8) + dimasklo
        rec['settable-readable-thresh-mask'] = (dimaskhi << 8) + dimasklo
        rec['sensor-units-1'] = sunit1
        # We'll need this in a minute.
        adf = (sunit1 & 0xc0) >> 6
        rec['analog-data-format'] = adf
        rec['rate-unit'] = (sunit1 & 0x7c) >> 3
        rec['modifier-unit-type'] = (sunit1 & 0x6) >> 1
        rec['unit-percentage'] = sunit1 & 0x1
        rec['base-unit'] = ipmievent.unittypelist[sunit2]
        rec['modifier-unit'] = ipmievent.unittypelist[sunit3]
        rec['linearization'] = lin
        rec['m'] = utos(((mtol & 0xc0) << 2) + m, 10)
        rec['raw-b'] = b
        rec['b'] = utos(((bacc & 0xc0) << 2) + b, 10)
        rec['raw-b-accuracy'] = bacc
        rec['tolerance'] = (mtol & 0x3f)
        rec['accuracy'] = (((acc & 0xf0) << 2) + (bacc & 0x3f))
        rec['accuracy-exponent'] = ((acc & 0xc) >> 3)
        rec['r-exponent'] = utos((rbexp & 0xf0) >> 4, 4)
        rec['b-exponent'] = utos(rbexp & 0xf, 4)
        rec['analog-characteristics'] = achar

        # The following entries need to be made signed/unsigned
        # by the contents of analog-data-format.
        
        rec['nominal-reading'] = check_if_format(adf, nread, 8)
        rec['normal-maximum'] = check_if_format(adf, nmax, 8)
        rec['normal-minimum'] = check_if_format(adf, nmin, 8)
        rec['sensor-maximum-reading'] = check_if_format(adf, smax, 8)
        rec['sensor-minimum-reading'] = check_if_format(adf, smin, 8)
        rec['upper-non-recoverable-thresh'] = check_if_format(adf, unonrecov, 8)
        rec['upper-critical-thresh'] = check_if_format(adf, ucrit, 8)
        rec['upper-non-critical-thresh'] = check_if_format(adf, unoncrit, 8)
        rec['lower-non-recoverable-thresh'] = check_if_format(adf, lnonrecov, 8)
        rec['lower-critical-thresh'] = check_if_format(adf, lcrit, 8)
        rec['lower-non-critical-thresh'] = check_if_format(adf, lnoncrit, 8)
        
        rec['positive-going-hysteresis'] = physt
        rec['negative-going-hysteresis'] = nhyst
        rec['id-type'] = idtype
        rec['id-string'] = idstr

    elif rtype == 0x2:
        try:
            sownid, solun, snum, entid, entinst, sinit, scap, stype, etype, amasklo, amaskhi, dmasklo, dmaskhi, dimasklo, dimaskhi, sunit1, sunit2, sunit3, sshare1, sshare2, physt, nhyst, rev1, rev2, rev3, oem, idtype = struct.unpack('<27B', data[5:32])
            idstr = data[32:]
        except:
            print "struct bombed, len = %d" % len(data[5:32])
            pbytes(data[5:])
            raise

        rec['sensor-owner-id'] = sownid
        rec['sensor-owner-lun'] = solun & 0x3
        rec['sensor-owner-channel'] = (solun & 0xf0) >> 4
        rec['sensor-number'] = snum
        rec['entity-id'] = entid
        if entid < len(ipmievent.entitytypelist):
            rec['entity-id-name'] = ipmievent.entitytypelist[entid]
        rec['entity-is-logical'] = entinst & 0x80
        rec['entity-instance'] = entinst & 0x7f
        rec['sensor-initialization'] = sinit
        rec['sensor-capabilities'] = scap
        rec['sensor-capability-hysteresis'] = (scap & 0x30) >> 4
        rec['sensor-capability-thresh'] = (scap & 0xc) >> 2
        rec['sensor-capability-events'] = scap & 0x3              
        rec['sensor-type'] = stype
        if stype < len(ipmievent.sensortypelist):
            rec['sensor-type-name'] = ipmievent.sensortypelist[stype]
        rec['event-type'] = etype
        if etype < len(ipmievent.sensortypelist):
            rec['event-type-name'] = ipmievent.eventtypelist[etype]
        rec['assert-event-mask'] = (amaskhi << 8) + amasklo
        rec['lower-thresh-mask'] = (amaskhi << 8) + amasklo
        rec['deassert-event-mask'] = (dmaskhi << 8) + dmasklo
        rec['upper-thresh-mask'] = (dmaskhi << 8) + dmasklo
        rec['discrete-reading-mask'] = (dimaskhi << 8) + dimasklo
        rec['settable-readable-thresh-mask'] = (dimaskhi << 8) + dimasklo
        rec['sensor-units-1'] = sunit1
        rec['analog-data-format'] = (sunit1 & 0xc0) >> 6
        rec['rate-unit'] = (sunit1 & 0x7c) >> 3
        rec['modifier-unit-type'] = (sunit1 & 0x6) >> 1
        rec['unit-percentage'] = sunit1 & 0x1
        rec['base-unit'] = ipmievent.unittypelist[sunit2]
        rec['modifier-unit'] = ipmievent.unittypelist[sunit3]
        rec['sensor-record-sharing'] = (sshare1 << 8) + sshare2
        rec['id-string-modifier-type'] = sshare1 & 0x30
        rec['share-count'] = sshare1 & 0x7
        rec['entity-instance-sharing'] = sshare2 & 0x80
        rec['id-string-modifier-offset'] = sshare2 & 0x7f
        rec['positive-going-hysteresis'] = physt
        rec['negative-going-hysteresis'] = nhyst
        rec['id-type'] = idtype
        rec['id-string'] = idstr

    elif rtype == 0x10:
        if sdrver == 0x90:
            # HP LPr, mainly
            try:
                dsaddr, busid, addrspan, aaddr1, aaddr2, dtype, dtmod, dloc, entid, oem, idtype = struct.unpack('<11B', data[5:16])
                idstr = data[16:]
            except:
                print "struct bombed in type 0x10, ver 0.9, len = %d" % len(data[5:])
                pbytes(data[5:])
                raise

            rec['device-address'] = dsaddr >> 1
            if busid & 0x40:
                rec['bus-type'] = 'private'
            else:
                rec['bus-type'] = 'IPMB'
            rec['device-type'] = rec['bus-type']
            rec['device-id'] = busid & 0x30 >> 4
            rec['private-bus-id'] = busid & 0x30 >> 4
            rec['address-span'] = addrspan & 0x7
            rec['device-access'] = aaddr1
            rec['firmware-revision'] = aaddr2
            rec['device-type-code'] = dtype
            rec['device-type-modifier'] = dtmod
            dinfo = ipmidevice.decode_device(dtype, dtmod)
            if dinfo and dinfo[0]:
                rec['device-type-name'] = dinfo[0]
            if dinfo[1]:
                rec['device-modifier-name'] = dinfo[1]
            rec['entity-instance'] = dloc >> 5
            rec['entity-id'] = dloc & 0x1f
            if (dloc & 0x1f) < len(ipmievent.entitytypelist):
                rec['entity-id-name'] = ipmievent.entitytypelist[dloc & 0x1f]
            rec['associated-entity-id'] = entid
            rec['oem'] = oem
            rec['id-type'] = idtype
            rec['id-string'] = idstr

        else: # version 1.0 and later
            try:
                daaddr, dsaddr, busid, addrspan, resv, dtype, dtmod, entid, entinst, oem, idtype = struct.unpack('<11B', data[5:16])
                idstr = data[16:]
            except:
                print "struct bombed in type 0x10, ver 0.9, len = %d" % len(data[5:])
                pbytes(data[5:])
                raise

            rec['device-access'] = daaddr >> 1
            rec['device-address'] = dsaddr >> 1
            rec['channel-number'] = ((dsaddr & 0x1) << 3) + (busid & 0xe0)
            rec['device-lun'] = (busid & 0x18) >> 3
            rec['private-bus-id'] = busid & 0x3
            if busid & 0x3:
                rec['bus-type'] = 'private'
            else:
                rec['bus-type'] = 'IPMB'
            rec['device-type'] = rec['bus-type']
            rec['address-span'] = addrspan & 0x7
            rec['device-type-code'] = dtype
            rec['device-type-modifier'] = dtmod
            dinfo = ipmidevice.decode_device(dtype, dtmod)
            if dinfo[0]:
                rec['device-type-name'] = dinfo[0]
            if dinfo[1]:
                rec['device-modifier-name'] = dinfo[1]
            rec['entity-id'] = entid
            if entid < len(ipmievent.entitytypelist):
                rec['entity-id-name'] = ipmievent.entitytypelist[entid]
            rec['entity-instance'] = entinst
            rec['oem'] = oem
            rec['id-type'] = idtype
            rec['id-string'] = idstr
            

    elif rtype == 0x11:
        try:
            daa, fruid, accessinfo, chan, rev1, devtype, devtmod, fruentid, fruentinst, oem, idtype = struct.unpack('<11B', data[5:16])
            idstr = data[16:]
        except:
            print "struct bombed in type 0x11, len = %d" % len(data[5:])
            pbytes(data[5:])
            raise
        rec['device-access'] = (daa >> 1)
        rec['device-address'] = fruid
        rec['access-info'] = accessinfo
        if accessinfo & 0x80:
            rec['device-type'] = 'logical'
        else:
            rec['device-type'] = 'physical'
        rec['device-lun'] = (accessinfo >> 3) & 0x3
        rec['private-bus-id'] = accessinfo & 0x7
        rec['channel-number'] = (chan >> 4)

        rec['device-type-code'] = devtype
        rec['device-type-modifier'] = devtmod
        dinfo = ipmidevice.decode_device(devtype, devtmod)
        if dinfo[0]:
            rec['device-type-name'] = dinfo[0]
        if dinfo[1]:
            rec['device-modifier-name'] = dinfo[1]
        rec['entity-id'] = fruentid
        if fruentid < len(ipmievent.entitytypelist):
            rec['entity-id-name'] = ipmievent.entitytypelist[fruentid]
        rec['entity-instance'] = fruentinst
        rec['oem'] = oem
        rec['id-type'] = idtype
        rec['id-string'] = idstr

    elif rtype == 0x12:
        try:
            devaddr, chan, pwrnotify, dcap, rev1, rev2, rev3, entid, entinst, oem, idtype = struct.unpack('<11B', data[5:16])
            idstr = data[16:]
        except:
            print "struct bombed in type 0x12, len = %d" % len(data[5:16])
            pbytes(data[5:])
            raise

        rec['device-address'] = devaddr
        rec['channel-number'] = (chan & 0xf)
        rec['state-initialization'] = pwrnotify
        rec['power-state-notification'] = (pwrnotify & 0xe0) >> 5
        rec['notify-system-power-state'] = (pwrnotify & 0x80) >> 7
        rec['notify-device-power-state'] = (pwrnotify & 0x40) >> 6
        if (pwrnotify & 0x20):
            rec['controller-type'] = 'static'
        else:
            rec['controller-type'] = 'dynamic'
        rec['global-initialization'] = pwrnotify & 0xf
        rec['log-init-agent-errors'] = pwrnotify & 0x8
        rec['log-init-agent-access'] = pwrnotify & 0x4
        if pwrnotify & 0x3 == 0x0:
            rec['event-generation'] = 'enabled'
        elif pwrnotify & 0x3 == 0x1:
            rec['event-generation'] = 'disabled'
        elif pwrnotify & 0x3 == 0x2:
            rec['event-generation'] = 'no-init'
        else:
            rec['event-generation'] = 'reserved'
        rec['device-capabilities-word'] = dcap
        dcaps = []
        if dcap & 0x80:
            dcaps.append("Chassis Device")
        if dcap & 0x40:
            dcaps.append("Bridge")
        if dcap & 0x20:
            dcaps.append("IPMB Event Generator")
        if dcap & 0x10:
            dcaps.append("IPMB Event Receiver")
        if dcap & 0x8:
            dcaps.append("FRU Inventory Device")
        if dcap & 0x4:
            dcaps.append("SEL Device")
        if dcap & 0x2:
            dcaps.append("SDR Repository Device")
        if dcap & 0x1:
            dcaps.append("Sensor Device")
        rec['device-capabilities'] = dcaps
        rec['entity-id'] = entid
        if entid < len(ipmievent.entitytypelist):
            rec['entity-id-name'] = ipmievent.entitytypelist[entid]
        rec['entity-instance'] = entinst
        rec['id-type'] = idtype
        rec['id-string'] = idstr

    return rec

# decode_sdr_repo_info: Decode 'Get SDR Repository Information' command
# into a dictionary

def decode_sdr_repo_info(data):
    rec = {}
    
    sdrver, reclo, rechi, freelo, freehi, addts, erasets, opsup = struct.unpack('<BBBBBLLB', data)
    rec['sdr-version'] = sdrver
    rec['record-count'] = ((rechi << 8) + reclo)
    rec['bytes-free'] = ((freehi << 8) + freelo)
    rec['last-record-added'] = addts
    rec['last-record-modified'] = erasets
    rec['operational-support'] = opsup
    rec['get-sdr-alloc-info'] = opsup & 0x1
    rec['reserve-sdr-repo'] = opsup & 0x2
    rec['partial-add-sdr'] = opsup & 0x4
    rec['delete-sdr'] = opsup & 0x8
    rec['update-modes'] = opsup & 0x60
    rec['overflow'] = opsup & 0x80
        
    return rec
    

# download_sdr: Download the entire SDR repository,
# return as a list of dictionaries from decode_sdr_record().

def download_sdr(sess):

    # Pull down the SDR Repo information
    #ret = sess.send(0x20, 0x0, 0x81, 0x0, 1, 0xa, 0x20, '')
    m = ipmimessage.IPMIMessage()
    m.quickload(0x20, 0x0, 0x81, 0x0, 1, 0xa, 0x20, '')
    rm = sess.sendmessage(m)
    ret = rm.data
    if ord(ret[0]) != 0x0:
        raise ValueError, "Couldn't get SDR respository info"
    #retcode, sdrver, reclo, rechi, freelo, freehi, addts, erasets, opsup = struct.unpack('<BBBBBBLLB', ret)

    rinfo = decode_sdr_repo_info(ret[1:])
    
    if rinfo['reserve-sdr-repo']:
        # Get a reservation
        # ret = sess.send(0x20, 0x0, 0x81, 0x0, 0x0, 0xA, 0x22, '')
        m = ipmimessage.IPMIMessage()
        m.quickload(0x20, 0x0, 0x81, 0x0, 0x0, 0xA, 0x22, '')
        rm = sess.sendmessage(m)
        ret = rm.data
        #bytes(ret)
        if ord(ret[0]) != 0x0:
            print "Couldn't reserve SDR (0x%x)" % ord(ret[0])
            sys.exit(2)
        resvid = (ord(ret[2])<<8) + ord(ret[1])
    else:
        resvid = 0

    # NOW grab the repo
    sdrs = {}
    seloff = 0
    while seloff != 0xffff:
        try:
            ret = ipmitransport.load_chunked_record(sess, 0x20, 0x0, 0x81, 0x0, 0x0, 0xA, 0x23, resvid, seloff, None)
            # [3:] slice rips off the completion code and next rec ID
            dr = decode_sdr_record(ret[3:])

            # little bit of special handling here.
            # Type 2 records define a 'share number', which is the number
            # of sensors that use the same info in the sdr. The share number
            # is a count of the number of records to duplicate.
            # We need to expand it and put it in the repo so applications
            # don't have to hunt it down for themselves.
            if dr['record-type'] == 0x2 and dr['share-count'] > 1:
                i = 0
                while i < dr['share-count']:
                    # construct new record -- this is OK since
                    # SDRs are only strings or integers, not objects
                    ndr = dr.copy()
                    apstr = ''
                    if dr['id-string-modifier-type']:
                        # alpha/ASCII
                        # this is fun ... offset 0-25 is A-Z,
                        # offset 26 is AA-AZ... yummy.
                        l = dr['id-string-modifier-offset'] / 26
                        r = dr['id-string-modifier-offset'] % 26
                        if l > 0:
                            apstr = apstr + chr(ord('A') + l)
                        apstr = apstr + chr(ord('A') + r)
                    else:
                        # numeric, easier
                        if dr['id-string-modifier-offset']:
                            apstr = str(i + dr['id-string-modifier-offset'])
                        else:
                            apstr = str(i + 1)
                    ndr['id-string'] = dr['id-string'] + apstr
                    ndr['sensor-number'] = dr['sensor-number'] + i
                    if dr['entity-instance-sharing']:
                          ndr['entity-instance'] = dr['entity-instance'] + i
                    sdrs[generate_sdr_key(ndr)] = ndr
                    i = i + 1

            else:
                sdrs[generate_sdr_key(dr)] = dr


                
            
            seloff = (ord(ret[2]) << 8) + ord(ret[1])
        except:
            print "Problem getting SDR entry"
            raise

##    if len(sdrs) != rinfo['record-count']:
##        raise ValueError, "INVARIANT: Didn't get all the SDRs?"

    return sdrs
