#
# Copyright (c) 2002 Doug White
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# $Id: ipmisel.py,v 1.1 2002/07/19 06:04:09 dwhite Exp $
#
# ipmisel -- Functions for handling the System Event Log
#
# Doug White

import struct
import time
import ipmievent
import ipmimessage
import ipmisdr
import ipmitransport

# download_sel: Download System Event Log into dictionary.
def download_sel(sess):
    sels = []
    dsel = []

    m = ipmimessage.IPMIMessage()
    m.quickload(0x20, 0x0, 0x20, 0x0, 0x0, 0xA, 0x42, '')
    rm = sess.sendmessage(m)

    resvid = (ord(rm.data[2])<<8) + ord(rm.data[1])

    seloff = 0
    while seloff != 0xffff:
        ret = ipmitransport.load_chunked_record(sess, 0x20, 0x0, 0x81, 0x0, 0x0, 0xA, 0x43, resvid, seloff, 16)
        seloff = (ord(ret[2]) << 8) + ord(ret[1])
        sels.append(ret[3:])

    # Important bits for interpretation:

    # Sensor type (36-3)
    # Event Type (section 36)
    # Event Data (dependent on type)

    # The Event Type dictates how the Event Data bytes
    # are parsed.
    # eoff is the Event Offset, selecting the specific transition
    # that caused the sensor to trigger.

    # IPMI 0.9 (emr 0x2) defined etypes 0xe6 and 0xe7 for
    # sensor-specific event codes. Translate into IPMI 1.5
    # sensor-specific codes. 0xe6 is a deassert and 0xe7 is
    # an assert. (these were broken out into the direction bit)

    for y in sels:
        rec = {}
        ridlo, ridhi, rectype = struct.unpack('<3B', y[:3])
        if rectype == 0x2:
            ridlo, ridhi, rectype, tstamp, gidlo, gidhi, emr, stype, snum, sdir, data1, data2, data3 = struct.unpack('<3BL9B', y )
            # Calculate event type and offset.
            etype = 0
            if emr == 0x2 and (sdir == 0xe6 or sdir == 0xe7 or sdir == 0xe8):
                etype = 0x6f
            else:
                etype = sdir & 0x7f
            eoff = data1 & 0xf
            
            rec['record-id'] = ((ridhi << 8) + ridlo)
            rec['record-type'] = rectype
            rec['timestamp'] = tstamp
            rec['generator-address'] = gidlo
            rec['generator-channel'] = (gidhi >> 4)
            rec['generator-lun'] = gidhi & 0x3
            rec['format-version'] = emr
            rec['sensor-type'] = stype
            if stype < len(ipmievent.sensortypelist):
                rec['sensor-type-name'] = ipmievent.sensortypelist[stype]
            rec['sensor-number'] = snum
            if (sdir & 0x80) >> 7:
                rec['event-direction'] = 'D'
            else:
                rec['event-direction'] = 'A'
            rec['event-type'] = etype
            if etype < len(ipmievent.sensortypelist):
                rec['event-type-name'] = ipmievent.eventtypelist[etype]
            rec['event-data-1'] = data1
            rec['event-data-2'] = data2
            rec['event-data-3'] = data3
            rec['event-offset'] = eoff

        dsel.append(rec)

    return dsel


# format_sel: Given output of download_sel(), format into a list of
# syslog-like lines. This has to pull down the SDRs to format.
def format_sels(sess, sels):
    sdrs = ipmisdr.download_sdr(sess)
    fsels = []


    for s in sels:
        etypestr = ""
        especstr = ""
        sensname = ""
        sensinfo = None
        if s['record-type'] != 0x2:
            continue
        snum = s['sensor-number']
        #(etypestr, especstr) = ipmievent.decode_event(etype, eoff, stype, data1, data2, data3)
        (etypestr, especstr) = ipmievent.decode_event(
            s['event-type'], s['event-offset'], s['sensor-type'],
            s['event-data-1'], s['event-data-2'], s['event-data-3'])
        sensinfo = sensinfo = sdrs.get((0x1, s['generator-address'], s['generator-lun'], s['sensor-number']))
        if not sensinfo:
            sensinfo = sdrs.get((0x2, s['generator-address'], s['generator-lun'], s['sensor-number']))
        if sensinfo:
            sensname = sensinfo['id-string'] + ' - '

        if not etypestr:
            etypestr = "0x%x" % s['event-type']
        if not especstr:
            especstr = "0x%x" % s['event-offset']


        # print "%s sel[%d]: %s - %s" % (time.ctime(tstamp), ((ridhi << 8) + ridlo), etypestr, especstr)
        fsels.append("%s sel[%d]: %s %s%s - %s" % (time.strftime('%b %d %H:%M:%S %Y', time.localtime(s['timestamp'])), s['record-id'], s['event-direction'], sensname, etypestr, especstr))

    return fsels
