#
# Copyright (c) 2002 Doug White
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# $Id: ipmismic.py,v 1.2 2002/05/14 23:07:03 dwhite Exp $
#
# ipmismic.py: SMIC system interface driver
#
# Doug White

import sys
sys.path.append('iport')
import time
from ioport import *

# SMIC Constants

# 0 = busywait
SMIC_SLEEP_INTERVAL = 0

SMIC_BASE_PORT = 0xca9
SMIC_DATA_PORT    =(SMIC_BASE_PORT)
SMIC_CONTROL_PORT =(SMIC_BASE_PORT+1)
SMIC_STATUS_PORT  =(SMIC_BASE_PORT+1)
SMIC_FLAGS_PORT   =(SMIC_BASE_PORT+2)

SMIC_RX_DATA_RDY       =0x80
SMIC_TX_DATA_RDY       =0x40
SMIC_SMI               =0x10
SMIC_EVT_ATN           =0x08
SMIC_SMS_ATN           =0x04
SMIC_BUSY              =0x01

SMIC_CC_SMS_GET_STATUS  =0x40
SMIC_CC_SMS_WR_START    =0x41
SMIC_CC_SMS_WR_NEXT     =0x42
SMIC_CC_SMS_WR_END      =0x43
SMIC_CC_SMS_RD_START    =0x44
SMIC_CC_SMS_RD_NEXT     =0x45
SMIC_CC_SMS_RD_END      =0x46

SMIC_SC_SMS_RDY         =0xc0
SMIC_SC_SMS_WR_START    =0xc1
SMIC_SC_SMS_WR_NEXT     =0xc2
SMIC_SC_SMS_WR_END      =0xc3
SMIC_SC_SMS_RD_START    =0xc4
SMIC_SC_SMS_RD_NEXT     =0xc5
SMIC_SC_SMS_RD_END      =0xc6


def smic_get_flags():
    return inb(SMIC_FLAGS_PORT)

def smic_get_status():
    return inb(SMIC_STATUS_PORT)

def smic_print_flags(state):
    print "Current SMIC flags: 0x%x " % state,
    if state & SMIC_RX_DATA_RDY:
        print "RX_DATA_RDY ",
    if state & SMIC_TX_DATA_RDY:
        print "TX_DATA_RDY ",
    if state & SMIC_SMI:
        print "SMI ",
    if state & SMIC_EVT_ATN:
        print "EVT_ATN ",
    if state & SMIC_SMS_ATN:
        print "SMS_ATN ",
    if state & SMIC_BUSY:
        print "BUSY ",
    print ''

def smic_wait_for_rx_ready():
    while not smic_get_flags() & SMIC_RX_DATA_RDY:
        time.sleep(SMIC_SLEEP_INTERVAL)

def smic_wait_for_tx_ready():
    while not smic_get_flags() & SMIC_TX_DATA_RDY:
        time.sleep(SMIC_SLEEP_INTERVAL)

def smic_wait_for_idle():
    while smic_get_flags() & SMIC_BUSY:
        time.sleep(SMIC_SLEEP_INTERVAL)

def smic_set_busy():
    outb(SMIC_FLAGS_PORT, SMIC_BUSY)

def smic_write_start(data):
    smic_wait_for_idle()
    outb(SMIC_CONTROL_PORT, SMIC_CC_SMS_WR_START)
    outb(SMIC_DATA_PORT, ord(data))
    smic_set_busy()
    smic_wait_for_idle()

def smic_write_next(data):
    smic_wait_for_idle()
    smic_wait_for_tx_ready()
    outb(SMIC_CONTROL_PORT, SMIC_CC_SMS_WR_NEXT)
    outb(SMIC_DATA_PORT, ord(data))
    smic_set_busy()
    smic_wait_for_idle()

def smic_write_end(data):
    smic_wait_for_idle()
    smic_wait_for_tx_ready()
    outb(SMIC_CONTROL_PORT, SMIC_CC_SMS_WR_END)
    outb(SMIC_DATA_PORT, ord(data))
    smic_set_busy()
    smic_wait_for_idle()

def smic_read_start():
    smic_wait_for_idle()
    smic_wait_for_rx_ready()
    outb(SMIC_CONTROL_PORT, SMIC_CC_SMS_RD_START)
    smic_set_busy()
    smic_wait_for_idle()
    return inb(SMIC_DATA_PORT)

def smic_read_next():
    smic_wait_for_idle()
    smic_wait_for_rx_ready()
    outb(SMIC_CONTROL_PORT, SMIC_CC_SMS_RD_NEXT)
    smic_set_busy()
    smic_wait_for_idle()
    return inb(SMIC_DATA_PORT)

def smic_read_end():
    smic_wait_for_idle()
    outb(SMIC_CONTROL_PORT, SMIC_CC_SMS_RD_END)
    smic_set_busy()
    smic_wait_for_idle()

def smic_write(data):

    outlen = 0
    
    if len(data) < 2:
           print "smic_write: Impossibly small buffer"
           return None

    smic_write_start(data[0])
    outlen = outlen + 1

    x = smic_get_status()
    if x != SMIC_SC_SMS_WR_START:
        print "smic_write: Error writing starting byte"
        return None

    for c in data[1:-1]:
        smic_write_next(c)
        outlen = outlen + 1
        if smic_get_status() != SMIC_SC_SMS_WR_NEXT:
            print "smic_write: Error writing byte"
            return None

    smic_write_end(data[-1])
    outlen = outlen + 1
    if smic_get_status() != SMIC_SC_SMS_WR_END:
        print "smic_write: Error writing ending byte"
        return None

    return outlen

def smic_read():
    data = ""

    data = data + chr(smic_read_start())

    if smic_get_status() == SMIC_SC_SMS_RD_END:
        return data

    if smic_get_status() != SMIC_SC_SMS_RD_START:
        print "smic_read: Error reading starting byte"
        return None

    data = data + chr(smic_read_next())
    while smic_get_status() == SMIC_SC_SMS_RD_NEXT:
        data = data + chr(smic_read_next())

    if smic_get_status() != SMIC_SC_SMS_RD_END:
        print "smic_read: Error finishing reading"
        return None

    smic_read_end()
    if smic_get_status() != SMIC_SC_SMS_RDY:
        print "smic_read: Error returning to RDY state"
        return None

    return data
    
    
    

        
           

         

    
