/*-
 * Copyright (c) 2002 Doug White
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * 
 */

/*
 *  smic-common.c
 *
 */

#include <sys/types.h>
#include <machine/cpufunc.h>
#include <machine/sysarch.h>
#include <stdio.h>
#include "smic.h"
#include "smic-common.h"

/* Examine flags register. */

u_int8_t 
smic_get_flags()
{
    return inb(SMIC_FLAGS_PORT);    
}

u_int8_t
smic_get_status()
{
    return inb(SMIC_STATUS_PORT);
}

void 
smic_print_flags(u_int8_t state)
{
    printf("Current SMIC flags: %#x : ", state);
    if(state & SMIC_RX_DATA_RDY) 
        printf("RX_DATA_RDY ");
    if(state & SMIC_TX_DATA_RDY)
        printf("TX_DATA_RDY ");
    if(state & SMIC_SMI)
        printf("SMI ");
    if(state & SMIC_EVT_ATN) 
        printf("EVT_ATN ");
    if(state & SMIC_SMS_ATN)
        printf("SMS_ATN ");
    if(state & SMIC_BUSY)
        printf("BUSY ");
    printf("\n");
}

/* Basic spin loops. */

void 
smic_wait_for_not_busy() 
{
    while(smic_get_flags() & SMIC_BUSY)
    {
        usleep(SMIC_SLEEP_USECS);
    }
}

void 
smic_wait_for_rx_ready()
{
    while(!(smic_get_flags() & SMIC_RX_DATA_RDY))
    {
        usleep(SMIC_SLEEP_USECS);
    }
}

void 
smic_wait_for_tx_ready()
{
    while(!(smic_get_flags() & SMIC_TX_DATA_RDY))
    {
        usleep(SMIC_SLEEP_USECS);
    }
}

void 
smic_wait_for_idle()
{
    while(smic_get_flags() & SMIC_BUSY)
    {
        usleep(SMIC_SLEEP_USECS);
    }
}

/* Flag manipulations. */

void 
smic_set_busy()
{
    outb(SMIC_FLAGS_PORT, SMIC_BUSY);
}

/* Basic functions for writing bytes. */

void 
smic_write_start(u_int8_t data)
{
    smic_wait_for_idle();
    outb(SMIC_CONTROL_PORT, SMIC_CC_SMS_WR_START);
    outb(SMIC_DATA_PORT, data);
    smic_set_busy();
    smic_wait_for_idle();
}

void 
smic_write_next(u_int8_t data)
{
    smic_wait_for_idle();
    smic_wait_for_tx_ready();
    outb(SMIC_CONTROL_PORT, SMIC_CC_SMS_WR_NEXT);
    outb(SMIC_DATA_PORT, data);
    smic_set_busy();
    smic_wait_for_idle();
}
void 
smic_write_end(u_int8_t data)
{
    smic_wait_for_idle();
    smic_wait_for_tx_ready();
    outb(SMIC_CONTROL_PORT, SMIC_CC_SMS_WR_END);
    outb(SMIC_DATA_PORT, data);
    smic_set_busy();
    smic_wait_for_idle();
}

/* Basic functions for doing reads. */

u_int8_t
smic_read_start()
{
    smic_wait_for_idle();
    smic_wait_for_rx_ready();
    outb(SMIC_CONTROL_PORT, SMIC_CC_SMS_RD_START);
    smic_set_busy();
    smic_wait_for_idle();
    return inb(SMIC_DATA_PORT);
}

u_int8_t
smic_read_next()
{
    smic_wait_for_idle();
    smic_wait_for_rx_ready();
    outb(SMIC_CONTROL_PORT, SMIC_CC_SMS_RD_NEXT);
    smic_set_busy();
    smic_wait_for_idle();
    return inb(SMIC_DATA_PORT);
}

void
smic_read_end()
{
    smic_wait_for_idle();
    outb(SMIC_CONTROL_PORT, SMIC_CC_SMS_RD_END);
    smic_set_busy();
    smic_wait_for_idle();
}

/* API read function. */

int
smic_write(u_int8_t* data, int len)
{
    int outlen=0;
    int x=0;
    u_int8_t* p=data;
    
    /* In case someone isn't paying attention */
    if(len < 2)
    {
        warnx("smic_write: Impossibly short message\n");
        return -1;
    }
    
    /* Start up write with first byte. */
    smic_write_start(*p);
    p++; outlen++;
    
    x = smic_get_status();
    if(x != SMIC_SC_SMS_WR_START) 
    {
        warnx("smic_write: Error writing starting byte (%d)\n", x);
        return -1;
    }
    
    /* Loop over rest of bytes. */
    for(; outlen < (len - 1); outlen++) 
    {
        smic_write_next(*p);
        p++;
        outlen++;
        x = smic_get_status();
    
        if(x != SMIC_SC_SMS_WR_NEXT) 
        {
            warnx("smic_write: Error writing byte (%d)\n", x);
            return -1;
        }

    }
    
    /* End with final byte. */
    smic_write_end(*p);
    x = smic_get_status();
    if(x != SMIC_SC_SMS_WR_END) 
    {
        warnx("smic_write: Error writing ending byte (%d)\n", x);
        return -1;
    }
    
    warnx("smic_write: Write return code %#x", inb(SMIC_DATA_PORT));

    outlen++;
    
    return outlen;
}

/* API read function. */

int
smic_read(u_int8_t* data, int len)
{
    int outlen=0;
    int x=0;
    u_int8_t* p=data;

    /* In case someone isn't paying attention */
    if(len < 2)
    {
        warnx("smic_read: Impossibly small buffer\n");
        return -1;
    }

    *p = smic_read_start();
    p++;
    outlen++;
    x = smic_get_status();
    if(x == SMIC_SC_SMS_RD_END)
    {
        /* only one byte to grab */
        return outlen;
    }
    
    if(x != SMIC_SC_SMS_RD_START) 
    {
        warnx("smic_read: Error reading starting byte (%d)\n", x);
        return -1;
    }

    while(outlen < len) 
    {
        *p = smic_read_next();
        p++;
        outlen++;
        x = smic_get_status();
        warnx("smic_read: Read byte %#x, outlen = %d, len = %d, status = %#x", *(p-1), outlen, len, x);
        if(x != SMIC_SC_SMS_RD_NEXT) break;
    }
    
    if(outlen < len && x != SMIC_SC_SMS_RD_END)
    {
        warnx("smic_read: Error reading byte (%d)\n", x);
        return -1;
    }
    
    smic_read_end();
    x = smic_get_status();
    
    if(x != SMIC_SC_SMS_RDY) 
    {
        warnx("smic_read: Error returning to RDY state (%d)\n", x);
        return -1;
    }
    
    return outlen;
}
